cdef class FileUtils(object):

    @staticmethod
    def readHashMap(fileName: str) -> dict:
        result = dict()
        file = open(fileName, "r")
        lines = file.readlines()
        file.close()
        for line in lines:
            items = line.strip().split()
            result[items[0]] = items[1]
        return result



import math
from functools import cmp_to_key

from Math.Vector cimport Vector


cdef class VectorizedDictionary(Dictionary):

    def __init__(self, comparator=None, fileName=None):
        """
        A constructor of VectorizedDictionary class which calls its super class Dictionary.
        """
        super().__init__(comparator)
        if fileName is not None:
            input_file = open(fileName, "r", encoding="utf8")
            lines = input_file.readlines()
            for line in lines:
                items = line.split()
                vector = Vector(0, 0)
                for i in range(1, len(items)):
                    vector.add(float(items[i]))
                current_word = VectorizedWord(items[0], vector)
                self.words.append(current_word)
            input_file.close()
            self.words.sort(key=cmp_to_key(self.comparator))

    cpdef addWord(self, VectorizedWord word):
        """
        The addWord method takes a VectorizedWord as an input and adds it to the words list.

        PARAMETERS
        ----------
        word : VectorizedWord
            VectorizedWord input.
        """
        self.words.append(word)

    cpdef VectorizedWord mostSimilarWord(self, str name):
        """
        The mostSimilarWord method takes a String name as an input, declares a maxDistance as -1 and creates a
        VectorizedWord word by getting the given name from words list. Then, it loops through the words list and if the
        current word is not equal to given input it calculates the distance between current word and given word by using
        dot product and updates the maximum distance. It then returns the result VectorizedWord which holds the most
        similar word to the given word.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        VectorizedWord
            VectorizedWord type result which holds the most similar word to the given word.
        """
        cdef double distance, max_distance
        cdef VectorizedWord result, word, current_word
        max_distance = -1
        result = None
        word = self.getWord(name)
        if word is None:
            return None
        for current_word in self.words:
            if current_word != word and isinstance(word, VectorizedWord):
                distance = word.getVector().dotProduct(current_word.getVector())
                if distance > max_distance:
                    max_distance = distance
                    result = current_word
        return result

    cdef makeComparator(self, VectorizedWord comparedWord):
        def compare(wordA, wordB):
            v = comparedWord.getVector()
            vA = wordA.getVector()
            vB = wordB.getVector()
            result1 = v.dotProduct(vB) / math.sqrt(v.dotProductWithSelf() * vB.dotProductWithSelf())
            result2 = v.dotProduct(vA) / math.sqrt(v.dotProductWithSelf() * vA.dotProductWithSelf())
            if result1 < result2:
                return -1
            elif result1 > result2:
                return 1
            else:
                return 0
        return compare

    cpdef list mostSimilarKWords(self, str name, int k):
        """
        The mostSimilarKWords method takes a String name and an integer k as inputs, and creates an list resultWords
        of type VectorizedWord and a VectorizedWord word by getting the given name from words list. Then, it loops
        through the words list and adds current word to the resultWords. It then sort resultWords list and if the size
        of the list is greater than given input k, it removes items from the ending. Then, it returns resultWords list.

        PARAMETERS
        ----------
        name : str
            String input.
        k : int
            Integer input.

        RETURNS
        -------
        list
            list result.
        """
        cdef list result_words
        cdef VectorizedWord word, current_word
        result_words = []
        word = self.getWord(name)
        if word is None:
            return result_words
        for current_word in self.words:
            result_words.append(current_word)
        result_words.sort(key=self.makeComparator(word))
        return result_words[0: k]

    cpdef list kMeansClustering(self, int iteration, int k):
        """
        The kMeansClustering method takes an integer iteration and k as inputs. K-means clustering aims to partition n
        observations into k clusters in which each observation belongs to the cluster with the nearest mean.

        PARAMETERS
        ----------
        iteration : int
            Integer input.
        k : int
            Integer input.

        RETURNS
        -------
        list
            list result which holds the k-means clustered words.
        """
        cdef list result, means
        cdef Vector v
        cdef int vector_size, i, j, max_cluster_index
        cdef double max_cluster_distance
        cdef VectorizedWord vectorized_word
        result = []
        means = []
        vector_size = self.words[0].getVector().size()
        for i in range(k):
            result.append([])
            v = Vector()
            v.initAllSame(vector_size, 0)
            means.append(v)
        for i in range(len(self.words)):
            result[i % k].append(self.words[i])
            means[i % k].add(self.words[i]).getVector()
        for i in range(k):
            means[i].divide(len(result[i]))
            means[i].divide(math.sqrt(means[i].dotProductWithSelf()))
        for i in range(iteration):
            for j in range(k):
                result[j].clear()
            for vectorized_word in self.words:
                max_cluster_distance = means[0].dotProduct(vectorized_word.getVector())
                max_cluster_index = 0
                for j in range(1, k):
                    cluster_distance = means[j].dotProduct(vectorized_word.getVector())
                    if cluster_distance > max_cluster_distance:
                        max_cluster_distance = cluster_distance
                        max_cluster_index = j
                result[max_cluster_index].append(vectorized_word)
            for j in range(k):
                means[j].clear()
                for word in result[j]:
                    means[j].add(word.getVector())
                means[j].divide(len(result[j]))
                means[j].divide(math.sqrt(means[j].dotProductWithSelf()))
        return result



import random


cdef class RandomArray(object):

    @staticmethod
    def normalizedArray(item_count: int) -> list:
        """
        The constructor of RandomNormalizedArray class gets an integer itemCount as an input. Creates a list of
        size itemCount and loops through each element of the list and initializes them with a random number by using
        random. Then, accumulates each element of the list and at the end divides each element with the resulting sum.

        Parameters
        ----------
        item_count : int
            input defining array size.
        """
        cdef int i
        cdef double total
        cdef list array
        total = 0.0
        array = []
        for i in range(item_count):
            array.append(random.uniform(0, 1))
            total += array[i]
        for i in range(item_count):
            array[i] /= total
        return array

    @staticmethod
    def indexArray(item_count: int, seed: int) -> list:
        """
        Returns an array of array indexes but shuffled. For example, if n is 4, the method returns an array of indexes
        0, 1, 2, 3 but shuffled.

        Parameters
        ----------
        item_count: int
            Number of elements in the array
        seed: int
            Seed used to shuffle array.

        Returns
        ----------
        list
            Index array
        """
        cdef list array
        array = [i for i in range(item_count)]
        random.seed(seed)
        random.shuffle(array)
        return array



from Dictionary.Word cimport Word


cdef class Dictionary:

    def __init__(self, comparator=None):
        """
        An empty constructor of Dictionary class.
        """
        self.words = []
        self.filename = ""
        if comparator is None:
            self.comparator = Dictionary.turkishLowerCaseComparator
        else:
            self.comparator = comparator

    @staticmethod
    def turkishLowerCaseComparator(wordA: Word, wordB: Word):
        """
        Compares two words in a case-sensitive manner.
        :param wordA: First word to compare.
        :param wordB: Second word to compare.
        :return: the value 0 if the wordA is equal to the wordB; a value less than 0 if this wordA is
        lexicographically less than wordB; and a value greater than 1 if this wordA is lexicographically greater
        than wordB.
        """
        cdef int i, first, second
        cdef str first_char, second_char
        LOWERCASE_LETTERS = "abcçdefgğhıijklmnoöprsştuüvyz"
        for i in range(min(len(wordA.getName()), len(wordB.getName()))):
            first_char = wordA.getName()[i:i + 1]
            second_char = wordB.getName()[i:i + 1]
            if first_char != second_char:
                if first_char in LOWERCASE_LETTERS and second_char not in LOWERCASE_LETTERS:
                    return -1
                elif first_char not in LOWERCASE_LETTERS and second_char in LOWERCASE_LETTERS:
                    return 1
                elif first_char in LOWERCASE_LETTERS and second_char in LOWERCASE_LETTERS:
                    first = LOWERCASE_LETTERS.index(first_char)
                    second = LOWERCASE_LETTERS.index(second_char)
                    if first < second:
                        return -1
                    elif first > second:
                        return 1
        if len(wordA.getName()) < len(wordB.getName()):
            return -1
        elif len(wordA.getName()) > len(wordB.getName()):
            return 1
        else:
            return 0

    @staticmethod
    def turkishIgnoreCaseComparator(wordA: Word, wordB: Word):
        """
        Compares two words in a case-insensitive manner.
        :param wordA: First word to compare.
        :param wordB: Second word to compare.
        :return: the value 0 if the wordA is equal to the wordB; a value less than 0 if this wordA is
        lexicographically less than wordB; and a value greater than 1 if this wordA is lexicographically greater
        than wordB.
        """
        cdef int i, first, second
        cdef str first_char, second_char
        IGNORE_CASE_LETTERS = "aAbBcCçÇdDeEfFgGğĞhHıIiİjJkKlLmMnNoOöÖpPrRsSşŞtTuUüÜvVyYzZ"
        for i in range(min(len(wordA.getName()), len(wordB.getName()))):
            first_char = wordA.getName()[i:i + 1]
            second_char = wordB.getName()[i:i + 1]
            if first_char != second_char:
                if first_char in IGNORE_CASE_LETTERS and second_char not in IGNORE_CASE_LETTERS:
                    return -1
                elif first_char not in IGNORE_CASE_LETTERS and second_char in IGNORE_CASE_LETTERS:
                    return 1
                elif first_char in IGNORE_CASE_LETTERS and second_char in IGNORE_CASE_LETTERS:
                    first = IGNORE_CASE_LETTERS.index(first_char)
                    second = IGNORE_CASE_LETTERS.index(second_char)
                    if first < second:
                        return -1
                    elif first > second:
                        return 1
        if len(wordA.getName()) < len(wordB.getName()):
            return -1
        elif len(wordA.getName()) > len(wordB.getName()):
            return 1
        else:
            return 0

    cpdef Word getWord(self, str name):
        """
        The getWord method takes a String name as an input and performs binary search within words list and assigns the
        result to integer variable middle. If the middle is greater than 0, it returns the item at index middle of words
        list, None otherwise.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        Word
            the item at found index of words {@link ArrayList}, null if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(name)
        middle = self.__getPosition(word)
        if middle >= 0:
            return self.words[middle]
        return None

    cpdef int getWordIndex(self, str name):
        """
        The getWordIndex method takes a String name as an input and performs binary search within words list and assigns
        the result to integer variable middle. If the middle is greater than 0, it returns the index middle, -1
        otherwise.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        int
            found index of words list, -1 if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(name)
        middle = self.__getPosition(word)
        if middle >= 0:
            return middle
        return -1

    cpdef removeWord(self, str name):
        """
        RemoveWord removes a word with the given name

        PARAMETERS
        ----------
        name : str
            Name of the word to be removed.
        """
        cdef int index
        index = self.getWordIndex(name)
        if index != -1:
            self.words.pop(index)

    cpdef int size(self):
        """
        The size method returns the size of the words list.

        RETURNS
        -------
        int
            The size of the words list.
        """
        return len(self.words)

    cpdef Word getWordWithIndex(self, int index):
        """
        The getWordWithIndex method which takes an index as an input and returns the value at given index of words list.

        PARAMETERS
        ----------
        index : int
            index to get the value.

        RETURNS
        -------
        Word
            The value at given index of words list.
        """
        return self.words[index]

    cpdef int longestWordSize(self):
        """
        The longestWordSize method loops through the words list and returns the item with the maximum word length.

        RETURNS
        -------
        int
            The item with the maximum word length.
        """
        cdef int max_length
        max_length = 0
        for word in self.words:
            if len(word.getName()) > max_length:
                max_length = len(word.getName())
        return max_length

    cpdef int __getPosition(self, Word word):
        """
        Checks if a given word exists in the dictionary by performing a binary search on the words array.
        :param word: Searched word
        :return: the index of the search word, if it is contained in the words array; otherwise,
        (-(insertion point) - 1). The insertion point is defined as the point at which the word would be inserted into
        the words array.
        """
        cdef int lo, hi, mid
        lo = 0
        hi = len(self.words) - 1
        while lo <= hi:
            mid = (lo + hi) // 2
            if self.comparator(self.words[mid], word) < 0:
                lo = mid + 1
            elif self.comparator(self.words[mid], word) > 0:
                hi = mid - 1
            else:
                return mid
        return -(lo + 1)

    cpdef int getWordStartingWith(self, str _hash):
        """
        The getWordStartingWith method takes a String hash as an input and performs binary search within words list and
        assigns the result to integer variable middle. If the middle is greater than 0, it returns the index middle,
        -middle-1 otherwise.

        PARAMETERS
        ----------
        _hash : str
            String input.

        RETURNS
        -------
        int
            Found index of words list, -middle-1 if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(_hash)
        middle = self.__getPosition(word)
        if middle < 0:
            return -middle - 1
        else:
            return middle

    def __repr__(self):
        return f"{self.words}"
